@@ -10,6 +10,11 @@ import array
 import io
 import sys
 import struct
+import logging
+
+
+log = logging.getLogger(__name__)
+
 
 # Apple's documentation of 'gvar':
 # https://developer.apple.com/fonts/TrueType-Reference-Manual/RM06/Chap6gvar.html
@@ -284,7 +289,8 @@ class table__g_v_a_r(DefaultTable.DefaultTable):
 		deltas_y, pos = GlyphVariation.decompileDeltas_(len(points), tupleData, pos)
 		deltas = [None] * numPointsInGlyph
 		for p, x, y in zip(points, deltas_x, deltas_y):
-				deltas[p] = (x, y)
+				if 0 <= p < numPointsInGlyph:
+					deltas[p] = (x, y)
 		return GlyphVariation(axes, deltas)
 
 	@staticmethod
@@ -514,17 +520,21 @@ class GlyphVariation(object):
 
 		MAX_RUN_LENGTH = 127
 		pos = 0
+		lastValue = 0
 		while pos < numPoints:
 			run = io.BytesIO()
 			runLength = 0
-			lastValue = 0
-			useByteEncoding = (points[pos] <= 0xff)
+			useByteEncoding = None
 			while pos < numPoints and runLength <= MAX_RUN_LENGTH:
 				curValue = points[pos]
 				delta = curValue - lastValue
-				if useByteEncoding and delta > 0xff:
+				if useByteEncoding is None:
+					useByteEncoding = 0 <= delta <= 0xff
+				if useByteEncoding and (delta > 0xff or delta < 0):
 					# we need to start a new run (which will not use byte encoding)
 					break
+				# TODO This never switches back to a byte-encoding from a short-encoding.
+				# That's suboptimal.
 				if useByteEncoding:
 					run.write(bytechr(delta))
 				else:
@@ -553,24 +563,39 @@ class GlyphVariation(object):
 			pos += 1
 		if numPointsInData == 0:
 			return (range(numPointsInGlyph), pos)
+
 		result = []
 		while len(result) < numPointsInData:
 			runHeader = byteord(data[pos])
 			pos += 1
 			numPointsInRun = (runHeader & POINT_RUN_COUNT_MASK) + 1
 			point = 0
-			if (runHeader & POINTS_ARE_WORDS) == 0:
-				for _ in range(numPointsInRun):
-					point += byteord(data[pos])
-					pos += 1
-					result.append(point)
+			if (runHeader & POINTS_ARE_WORDS) != 0:
+				points = array.array("H")
+				pointsSize = numPointsInRun * 2
 			else:
-				for _ in range(numPointsInRun):
-					point += struct.unpack(">H", data[pos:pos+2])[0]
-					pos += 2
-					result.append(point)
-		if max(result) >= numPointsInGlyph:
-			raise TTLibError("malformed 'gvar' table")
+				points = array.array("B")
+				pointsSize = numPointsInRun
+			points.fromstring(data[pos:pos+pointsSize])
+			if sys.byteorder != "big":
+				points.byteswap()
+
+			assert len(points) == numPointsInRun
+			pos += pointsSize
+
+			result.extend(points)
+
+		# Convert relative to absolute
+		absolute = []
+		current = 0
+		for delta in result:
+			current += delta
+			absolute.append(current)
+		result = absolute
+		del absolute
+
+		if max(result) >= numPointsInGlyph or min(result) < 0:
+			log.warning("point number out of range in 'gvar' table")
 		return (result, pos)
 
 	def compileDeltas(self, points):
@@ -696,14 +721,19 @@ class GlyphVariation(object):
 			numDeltasInRun = (runHeader & DELTA_RUN_COUNT_MASK) + 1
 			if (runHeader & DELTAS_ARE_ZERO) != 0:
 				result.extend([0] * numDeltasInRun)
-			elif (runHeader & DELTAS_ARE_WORDS) != 0:
-				for _ in range(numDeltasInRun):
-					result.append(struct.unpack(">h", data[pos:pos+2])[0])
-					pos += 2
 			else:
-				for _ in range(numDeltasInRun):
-					result.append(struct.unpack(">b", data[pos:pos+1])[0])
-					pos += 1
+				if (runHeader & DELTAS_ARE_WORDS) != 0:
+					deltas = array.array("h")
+					deltasSize = numDeltasInRun * 2
+				else:
+					deltas = array.array("b")
+					deltasSize = numDeltasInRun
+				deltas.fromstring(data[pos:pos+deltasSize])
+				if sys.byteorder != "big":
+					deltas.byteswap()
+				assert len(deltas) == numDeltasInRun
+				pos += deltasSize
+				result.extend(deltas)
 		assert len(result) == numDeltas
 		return (result, pos)
 
